mapping_config = {
    'rows': 10,
    'cols': 10,
    'empty_cell_color': '#ff00ff',
    'color_mapping':
    {
        'active_monster': '#ff00ff',
        'enemy_active_monster': '#ff00ff',
        'active_stats_modifier_monster': '#ff00ff',
        'enemy_active_stats_modifier_monster': '#ff00ff',
        'p1_field': '#ff00ff',
        'p2_field': '#ff00ff',
    }
}

cell_data = [
    {
        'row': 1,
        'col': 1,
        'type': 'active_monster',
        'name': 'Pikachu',
        'details': '"stats":{"atk":254,"def":206,"spa":134,"spd":174,"spe":150}',
        'actions': ['Thunder', 'Thunder Wave', 'Thundershock', 'Agility', ],
    }
]

class MapAnalyzer():
    def __init__(self, mapping_config):

        self.mapping_config = mapping_config
        self.cell_data = {}


    def process_frame(self, frame_key, cell_data):

        # resetooo
        self.cell_data[frame_key] = cell_data



import json
from .text_tracker import TextTracker
import cv2


DEFAULT_IGNORE_TRIGGER_THRESHOLDS = 0.05
DEFAULT_OVERRIDDE_RECT_THRESHOLDS = 0.93
DEFAULT_OCR_CONFIG =  {
    "always_use_list": False,
    "language_configs": [
        {
            "language": "eng",
            "use_easy_orc": False,
            "invert": False,
            "psm7": False,
        },
    ]
}


class TextConfigGenerator:

    def __init__(self, default_ocr_config=None):
        # only used for exluding from pairs
        self.labels_to_track = set()
        self.tracking_configurations = {}
        self.default_ocr_config = default_ocr_config

    def add_label_to_configuration(self, images, label_to_track, ocr_rect, other_ocr_rects=None, true_label=None, source_group=None, assisted_label_rect=None, ignore_if_labels_exists=None, ignore_if_regex_triggered=None, ignore_trigger_threshold=0.0, ignore_trigger_rects=None,
                override_rect_threshold=0.0, override_rects=None, ocr_config=None, other_ocr_configs=None, frames_to_skip=None, override_style=None):

        if true_label is None:
            true_label = label_to_track

        # true labels only. secondaries not included.
        self.labels_to_track.add(label_to_track)
        config = {
            "label": true_label,
            "position": "fixed",
            "ocr_rect": ocr_rect,
        }

        if frames_to_skip is None:
            frames_to_skip = 0

        if override_style in [None, 'base_style', 'custom']:
            config['override_style'] = override_style
        else:
            config['override_style'] = None

        config['frames_to_skip'] = frames_to_skip

        if other_ocr_rects is not None:
            config['other_ocr_rects'] = other_ocr_rects

        if assisted_label_rect is not None:
            config['assisted_label_rect'] = assisted_label_rect

        # add
        if source_group is not None:
            config['source_group'] = source_group

        if ignore_if_labels_exists is not None:
            config['ignore_if_labels_exists'] = ignore_if_labels_exists

        if ignore_if_regex_triggered is not None:
            config['ignore_if_regex_triggered'] = ignore_if_regex_triggered

        if ignore_trigger_rects is not None:
            ignore_config = {}
            ignore_config['rects'] = TextTracker.get_min_max_colors_for_regions_in_images(images, ignore_trigger_rects)
            ignore_config['threshold'] = ignore_trigger_threshold
            config['ignore_trigger_rects'] = ignore_config


        if override_rects is not None:
            config['override_rects'] = TextTracker.get_min_max_colors_for_regions_in_images(images, override_rects)
            config['threshold'] = ignore_trigger_threshold

        if ocr_config is not None:
            config['ocr_config'] = ocr_config
        elif self.default_ocr_config is not None:
            config['ocr_config'] = self.default_ocr_config

        if other_ocr_configs is not None:
            config['other_ocr_configs'] = other_ocr_configs

        self.tracking_configurations[label_to_track] = config

    def get_generated_config(self):
        return {
            "labels_to_track": list(self.labels_to_track),
            "tracking_configurations": self.tracking_configurations,
        }

    def get_generated_config_as_json(self):
        return json.dumps(self.get_generated_config(), indent=4)







    @staticmethod
    def config_generator_from_mapping(mapping_json):
      label_keys = mapping_json["config_labels"]
      cfg_generator = TextConfigGenerator()

      for label_key in label_keys:
        label_to_track = label_key
        ocr_rect = mapping_json["ocr_rects"][label_key]
        assisted_label_rect = ocr_rect

        other_ocr_rects = None
        if "other_ocr_rects" in mapping_json:
          if label_to_track in mapping_json["other_ocr_rects"]:
            other_ocr_rects = mapping_json["other_ocr_rects"][label_to_track]

        if "assist_rects" in mapping_json:
          if label_to_track in mapping_json["assist_rects"]:
            assisted_label_rect = mapping_json["assist_rects"][label_to_track]

        ignore_trigger_threshold = DEFAULT_IGNORE_TRIGGER_THRESHOLDS
        override_rect_threshold = DEFAULT_OVERRIDDE_RECT_THRESHOLDS

        ignore_trigger_rects = None
        override_rects = None

        config_creation_images = []
        for image in mapping_json["images"][label_to_track]:
          if isinstance(image, str):
            config_creation_images.append(cv2.imread(image))
          else:
            config_creation_images.append(image)

        ocr_config = DEFAULT_OCR_CONFIG

        ignore_if_labels_exists = None
        ignore_if_regex_triggered = None

        true_label = label_to_track
        if "true_label_mapping" in mapping_json:
          if label_to_track in mapping_json["true_label_mapping"]:
            true_label = mapping_json["true_label_mapping"][label_to_track]


        source_group = None
        if "source_groups" in mapping_json:
          if label_to_track in mapping_json["source_groups"]:
            source_group = mapping_json["source_groups"][label_to_track]

        if "ignore_labels" in mapping_json:
          if label_to_track in mapping_json["ignore_labels"]:
            ignore_if_labels_exists = mapping_json["ignore_labels"][label_to_track]

        if "ignore_regexes" in mapping_json:
          if label_to_track in mapping_json["ignore_regexes"]:
            ignore_if_regex_triggered = mapping_json["ignore_regexes"][label_to_track]

        if "ignore_thresholds" in mapping_json:
          if label_to_track in mapping_json["ignore_thresholds"]:
            ignore_trigger_threshold = mapping_json["ignore_thresholds"][label_to_track]

        if "override_thresholds" in mapping_json:
          if label_to_track in mapping_json["override_thresholds"]:
            override_rect_threshold = mapping_json["override_thresholds"][label_to_track]

        if "override_rects" in mapping_json:
          if label_to_track in mapping_json["override_rects"]:
            override_rects = mapping_json["override_rects"][label_to_track]

        if "ignore_trigger_rects" in mapping_json:
          if label_to_track in mapping_json["ignore_trigger_rects"]:
            ignore_trigger_rects = mapping_json["ignore_trigger_rects"][label_to_track]


        if "ocr_configs" in mapping_json:
          if label_to_track in mapping_json["ocr_configs"]:
            ocr_config = mapping_json["ocr_configs"][label_to_track]

        other_ocr_configs = None
        if "other_ocr_configs" in mapping_json:
          if label_to_track in mapping_json["other_ocr_configs"]:
            other_ocr_configs = mapping_json["other_ocr_configs"][label_to_track]

        frames_to_skip = None
        if "frames_to_skip" in mapping_json:
          if label_to_track in mapping_json["frames_to_skip"]:
            frames_to_skip = mapping_json["frames_to_skip"][label_to_track]

        override_style = None
        if "override_styles" in mapping_json:
          if label_to_track in mapping_json["override_styles"]:
            override_style = mapping_json["override_styles"][label_to_track]

        cfg_generator.add_label_to_configuration(config_creation_images, label_to_track, ocr_rect, other_ocr_rects=other_ocr_rects, true_label=true_label, source_group=source_group, assisted_label_rect=assisted_label_rect, ignore_if_labels_exists=ignore_if_labels_exists,
                                              ignore_if_regex_triggered=ignore_if_regex_triggered,
                                              ignore_trigger_threshold=ignore_trigger_threshold, ignore_trigger_rects=ignore_trigger_rects,
                        override_rect_threshold=override_rect_threshold, override_rects=override_rects, ocr_config=ocr_config, other_ocr_configs=other_ocr_configs, frames_to_skip=frames_to_skip, override_style=override_style)
      return cfg_generator



#Patent Pending

# Used to respond to events of text appearing on screen.
#    "apply_tesseract": False,

RECT_KEY = 'rect'
OCR_RECT_KEY = 'ocr_rect'
RECTS_KEY = 'rects'
ASSISTED_LABEL_RECT_KEY = 'assisted_label_rect'
THRESHOLD_KEY = 'threshold'
OCR_CONFIG_KEY = 'ocr_config'
LABELS_TO_TRACK_KEY = 'labels_to_track'
OVERRIDE_RECTS_KEY = 'override_rects'
IGNORE_IF_TRIGGERS_KEY = 'ignore_if_triggers'
IGNORE_IF_LABELS_EXIST_KEY = 'ignore_if_labels_exists'
IGNORE_IF_REGEX_TRIGGERED_KEY = 'ignore_if_regex_triggered'
REQUIRED_REGEX_TRIGGERED_KEY = 'ignore_if_regex_triggered'
TRACKING_CONFIGURATIONS_KEY = 'tracking_configurations'

import os
import re
REJECT_DUPLICATE_MESSAGES = bool(int(os.environ.get('REJECT_DUPLICATE_MESSAGES', 1)))
SAVE_MESSAGE_FRAMES = bool(int(os.environ.get('SAVE_MESSAGE_FRAMES', 1)))
ACCEPTED_MESSAGE_DIR = './accepted_message_frames'
REJECTED_MESSAGE_DIR = './rejected_message_frames'
DRAW_FRAMES_DIR = 'draw_frames'

import uuid

import io
from PIL import Image

try:
    from StringIO import StringIO ## for Python 2
except ImportError:
    from io import StringIO ## for Python 3

class TextTracker():
    def __init__(self, config, callback=None, pool=None, custom_override_detector=None):

        self.config = config
        self.futures = []
        self.pool = pool
        self.callback = callback


        self.occurences  = {}
        self.assisted_occurences = {}
        self.ignored_occurences = {}
        self.processed_text = {}
        self.rejected_text = {}

        self.all_accepted_messages = []
        self.all_rejected_messages = []
        self.messages_in_order = []
        self.text_counts = {}

        # For local storage
        self.stored_messages_for_local = {}
        self.last_frame_mapping = {}
        self.messages_frame_count = {}
        self.current_frame_number = 0
        self.custom_override_detector = custom_override_detector

        for label_to_track in self.config[LABELS_TO_TRACK_KEY]:
            self.messages_frame_count[label_to_track] = 0

    def apply_async_pool(self, pool):
        self.pool = pool

    def apply_async_pool(self, pool):
        self.pool = pool

    def process_frame(self, frame_key, frame, label_and_boxes):

        labels_to_track = self.config[LABELS_TO_TRACK_KEY]
        # first pass is only for checking if frame exists or needs to be added.

        assist_labels = []

        # resetooo
        should_ignore_override_set = set()
        self.occurences  = {}
        self.assisted_occurences = {}
        self.ignored_occurences = {}
        self.processed_text = {}
        self.rejected_text = {}

        # used to know whether or not to consider variations
        # of the same label. pokemon_sword_message vs twitch_pokemon_sword
        # if one succeeds, the other doesnt need consideration
        processed_true_labels = set()

        active_source_group = None

        #this method will check pokemon_message, but not consider twitch_pokemon_message until override checks
        for label_to_track in labels_to_track:
            did_assist_in_detection = False

            if label_to_track not in label_and_boxes:
                continue

            text_track_configuration = self.config[TRACKING_CONFIGURATIONS_KEY][label_to_track]

            if "source_group" in text_track_configuration:
                active_source_group = text_track_configuration["source_group"]
                break


        # used to determine group before processing active

        for label_to_track in labels_to_track:
            did_assist_in_detection = False

            if label_to_track in label_and_boxes:
                continue
            if label_to_track not in self.config[TRACKING_CONFIGURATIONS_KEY]:
                print('unsupported label', label_to_track)
                continue

            text_track_configuration = self.config[TRACKING_CONFIGURATIONS_KEY][label_to_track]

            source_group = None

            if "source_group" in text_track_configuration:
                source_group = text_track_configuration["source_group"]

            # early exit for group barrier
            if source_group is not None and active_source_group is not None and source_group != active_source_group:
                continue


            text_track_rect = text_track_configuration[OCR_RECT_KEY]
            ocr_config = None
            if OCR_CONFIG_KEY in text_track_configuration:
                ocr_config = text_track_configuration[OCR_CONFIG_KEY]


            text_frame = self.get_subframe_from_frame(text_track_rect, frame)

            # check for ignore triggers
            if IGNORE_IF_TRIGGERS_KEY in text_track_configuration:
                override_data = text_track_configuration[IGNORE_IF_TRIGGERS_KEY]
                base_override_threshold = 0.05
                if THRESHOLD_KEY in override_data:
                    base_override_threshold = override_data[THRESHOLD_KEY]

                # check for overrides
                should_ignore_override, _ = TextTracker.text_override_in_frame(frame, override_data[RECTS_KEY], base_override_threshold)

                if should_ignore_override:
                    if frame_key not in self.ignored_occurences:
                        self.ignored_occurences[frame_key] = []

                    if label_to_track not in self.ignored_occurences[frame_key]:
                        self.ignored_occurences[frame_key].append(label_to_track)

                    if frame_key not in self.occurences:
                        self.occurences[frame_key] = []

                    if label_to_track not in self.text_counts:
                        self.text_counts[label_to_track] = 0

                    if label_to_track not in self.occurences[frame_key]:
                        self.occurences[frame_key].append(label_to_track)
                    self.text_counts[label_to_track] += 1

                    should_ignore_override_set.add(label_to_track)


            if OVERRIDE_RECTS_KEY not in text_track_configuration or len(text_track_configuration[OVERRIDE_RECTS_KEY]) == 0:
                continue

            override_style = None
            if "override_style" in text_track_configuration:
                override_style = text_track_configuration["override_style"]


            base_threshold = 0.95
            if THRESHOLD_KEY in text_track_configuration:
                base_threshold = text_track_configuration[THRESHOLD_KEY]
            base_threshold = 0.95

            did_assist_in_detection = False
            if override_style == "custom" and self.custom_override_detector is not None:
                did_assist_in_detection = self.custom_override_detector(frame, text_track_configuration[OVERRIDE_RECTS_KEY], base_threshold)

            else:
                # check for overrides
                override_for_text, _ = TextTracker.text_override_in_frame(frame, text_track_configuration[OVERRIDE_RECTS_KEY], base_threshold)

                #an assist did happen
                did_assist_in_detection = override_for_text

            if did_assist_in_detection:
                assist_labels.append(label_to_track)

                # add rect to labels and boxes.
                rect_for_assist = text_track_configuration[OCR_RECT_KEY]
                if ASSISTED_LABEL_RECT_KEY in text_track_configuration:
                    rect_for_assist = text_track_configuration[ASSISTED_LABEL_RECT_KEY]

                label_and_boxes[label_to_track] = [rect_for_assist]
                if frame_key not in self.assisted_occurences:
                    self.assisted_occurences[frame_key] = []

                if label_to_track not in self.assisted_occurences[frame_key]:
                    self.assisted_occurences[frame_key].append(label_to_track)

        # second pass
        for label_to_track in labels_to_track:

            text_track_configuration = self.config[TRACKING_CONFIGURATIONS_KEY][label_to_track]

            source_group = None

            if "source_group" in text_track_configuration:
                source_group = text_track_configuration["source_group"]

            if source_group is not None and active_source_group is not None and source_group != active_source_group:
                continue

            active_source_group = source_group

            true_tracking_label = text_track_configuration["label"]

            if true_tracking_label in processed_true_labels:
                continue

            if label_to_track not in label_and_boxes:
                continue

            if label_to_track in should_ignore_override_set:
                continue

            if label_to_track not in self.config[TRACKING_CONFIGURATIONS_KEY]:
                print('unsupported label', label_to_track)
                continue

            required_skip_frames = 0
            if "frames_to_skip" in text_track_configuration:
                required_skip_frames = text_track_configuration["frames_to_skip"]


            if self.messages_frame_count[label_to_track] == 0 or required_skip_frames == 0:
                pass
            elif self.messages_frame_count[label_to_track] - self.current_frame_number < required_skip_frames:
                    continue

            self.messages_frame_count[label_to_track] = self.current_frame_number

            ignore_based_on_label = False
            if IGNORE_IF_LABELS_EXIST_KEY in text_track_configuration:
                for ignore_label in text_track_configuration[IGNORE_IF_LABELS_EXIST_KEY]:
                    if ignore_label in label_and_boxes:
                        ignore_based_on_label = True
            if ignore_based_on_label:
                if frame_key not in self.ignored_occurences:
                    self.ignored_occurences[frame_key] = []

                if label_to_track not in self.ignored_occurences[frame_key]:
                    self.ignored_occurences[frame_key].append(label_to_track)
                continue

            if frame_key not in self.occurences:
                self.occurences[frame_key] = []

            if label_to_track not in self.text_counts:
                self.text_counts[label_to_track] = 0

            if label_to_track not in self.occurences[frame_key]:
                self.occurences[frame_key].append(label_to_track)
            self.text_counts[label_to_track] += 1

            all_ocr_rects_and_configs = []

            text_track_rect = text_track_configuration[OCR_RECT_KEY]

            ocr_config = {}
            if OCR_CONFIG_KEY in text_track_configuration:
                ocr_config = text_track_configuration[OCR_CONFIG_KEY]

            all_ocr_rects_and_configs.append((None, text_track_rect, ocr_config))

            if "other_ocr_rects" in text_track_configuration:
                for ocr_key in text_track_configuration["other_ocr_rects"]:
                    ocr_config_to_use = ocr_config
                    if "other_ocr_configs" in text_track_configuration and ocr_key in text_track_configuration["other_ocr_configs"]:
                        ocr_config_to_use = text_track_configuration["other_ocr_configs"][ocr_key]
                    all_ocr_rects_and_configs.append((ocr_key, text_track_configuration["other_ocr_rects"][ocr_key], ocr_config))

            for ocr_rect_and_config in all_ocr_rects_and_configs:
                ocr_key, text_track_rect, ocr_config = ocr_rect_and_config


                message_frame = self.get_subframe_from_frame(text_track_rect, frame)
                retval, buffer = cv2.imencode('.png', message_frame)

                text_img_bytes = np.array(buffer).tostring()

                last_frame_key = '%s_%s' % (label_to_track, ocr_key)
                last_message_bytes = self.update_and_get_last_message_frame(text_img_bytes, last_frame_key)
                message_similarities = last_message_bytes is not None


                restricted_regex = None
                if IGNORE_IF_REGEX_TRIGGERED_KEY in text_track_configuration:
                    restricted_regex = text_track_configuration[IGNORE_IF_REGEX_TRIGGERED_KEY]

                if last_message_bytes is not None:
                    last_message_frame = np.array(Image.open(io.BytesIO(last_message_bytes)))

                    message_similarities = self.calculate_similar_frame_ratio(last_message_frame, message_frame)

                    if REJECT_DUPLICATE_MESSAGES and( last_message_bytes is not None and message_similarities >= 0.95):
                        print('Rejecting frame for duplicate message')
                        self.save_message_img(message_frame, accepted=False)
                    else:
                        self.save_message_img(message_frame, accepted=True)
                        if self.callback is not None:
                            self.callback(label_to_track, message_frame, ocr_config, ocr_key)
                        if self.pool is not None:
                            future = self.pool.apply_async(self.extract_english_japanese_text, [frame_key, message_frame, ocr_config, label_to_track, ocr_key, restricted_regex], callback=self.construct_pokemon_info_page_1)
                            self.futures.append(future)

                else:
                    self.save_message_img(message_frame, accepted=True)
                    if self.callback is not None:
                        self.callback(label_to_track, message_frame, ocr_config, ocr_key)
                    if self.pool is not None:
                        future = self.pool.apply_async(self.extract_english_japanese_text, [frame_key, message_frame, ocr_config, label_to_track, ocr_key, restricted_regex], callback=self.construct_pokemon_info_page_1)

        self.current_frame_number += 1
        return label_and_boxes, assist_labels

    def update_and_get_last_message_frame(self, current_frame, label):
        last_frame = None
        if label in self.last_frame_mapping:
            last_frame = self.last_frame_mapping[label]
        self.last_frame_mapping[label] = current_frame

        return last_frame

    def clear_last_message_frame(self, label):
        self.last_frame_mapping[label] = None

    # used for quick sanity checking
    def get_labels_and_ignore_override_ranges(self):
        labels_and_overrides = {}

        labels_to_track = self.config[LABELS_TO_TRACK_KEY]
        for label_to_track in labels_to_track:
            text_track_configuration = self.config[TRACKING_CONFIGURATIONS_KEY][label_to_track]

            # check for ignore triggers
            if IGNORE_IF_TRIGGERS_KEY not in text_track_configuration:
                continue

            override_configs = text_track_configuration[IGNORE_IF_TRIGGERS_KEY]
            max_red_gap = 0
            max_green_gap = 0
            max_blue_gap = 0

            labels_and_overrides[label_to_track] = {}
            labels_and_overrides[label_to_track]["rect_gaps"] = []

            for override_config in override_configs:
                red_gap = override_config["min_max_red"][1] - override_config["min_max_red"][0]
                green_gap = override_config["min_max_green"][1] - override_config["min_max_green"][0]
                blue_gap = override_config["min_max_blue"][1] - override_config["min_max_blue"][0]

                max_red_gap = max(max_red_gap, red_gap)
                max_green_gap = max(max_green_gap, green_gap)
                max_blue_gap = max(max_blue_gap, blue_gap)

                labels_and_overrides[label_to_track]["rect_gaps"].append((red_gap, green_gap, blue_gap))

            labels_and_overrides[label_to_track]["max_red_gap"] = max_red_gap
            labels_and_overrides[label_to_track]["max_green_gap"] = max_green_gap
            labels_and_overrides[label_to_track]["max_blue_gap"] = max_blue_gap

        return labels_and_overrides


    def get_labels_and_override_ranges(self):
        labels_and_overrides = {}

        labels_to_track = self.config[LABELS_TO_TRACK_KEY]
        for label_to_track in labels_to_track:

            text_track_configuration = self.config[TRACKING_CONFIGURATIONS_KEY][label_to_track]

            if OVERRIDE_RECTS_KEY not in text_track_configuration:
                continue

            override_configs = text_track_configuration[OVERRIDE_RECTS_KEY]
            max_red_gap = 0
            max_green_gap = 0
            max_blue_gap = 0

            labels_and_overrides[label_to_track] = {}
            labels_and_overrides[label_to_track]["rect_gaps"] = []

            for override_config in override_configs:
                red_gap = override_config["min_max_red"][1] - override_config["min_max_red"][0]
                green_gap = override_config["min_max_green"][1] - override_config["min_max_green"][0]
                blue_gap = override_config["min_max_blue"][1] - override_config["min_max_blue"][0]

                max_red_gap = max(max_red_gap, red_gap)
                max_green_gap = max(max_green_gap, green_gap)
                max_blue_gap = max(max_blue_gap, blue_gap)

                labels_and_overrides[label_to_track]["rect_gaps"].append((red_gap, green_gap, blue_gap))

            labels_and_overrides[label_to_track]["max_red_gap"] = max_red_gap
            labels_and_overrides[label_to_track]["max_green_gap"] = max_green_gap
            labels_and_overrides[label_to_track]["max_blue_gap"] = max_blue_gap

        return labels_and_overrides


    def get_warnings_for_overrides(self, threshold=100):
        ignored_warnings = []
        override_warnings = []

        ignored_labels_and_overrides = self.get_labels_and_ignore_override_ranges()
        labels_and_overrides = self.get_labels_and_override_ranges()

        for label in ignored_labels_and_overrides:
            override_data = ignored_labels_and_overrides[label]
            max_gap = max(override_data["max_red_gap"], override_data["max_red_gap"], override_data["max_red_gap"])
            if max_gap >= threshold:
                print('warning for: ', label, 'with max gap', max_gap, 'with data:', override_data)
                ignored_warnings.append((label, max_gap))

        for label in labels_and_overrides:
            override_data = labels_and_overrides[label]
            max_gap = max(override_data["max_red_gap"], override_data["max_red_gap"], override_data["max_red_gap"])
            if max_gap >= threshold:
                print('warning for: ', label, 'with max gap', max_gap, 'with data:', override_data)
                override_warnings.append((label, max_gap))

        return override_warnings, ignored_warnings

    @staticmethod
    def get_subframe_from_frame(rect, source_frame):
        h, w, c = source_frame.shape
        x1, y1, x2, y2 = rect
        x1 = min(max(x1, 0), w)
        x2 = min(max(x2, 0), w)
        y1 = min(max(y1, 0), h)
        y2 = min(max(y2, 0), h)
        crop_img = source_frame[y1:y2, x1:x2]
        return crop_img

    def add_messages_to_session(self, messages, label):
        if label not in self.stored_messages_for_local:
            self.stored_messages_for_local[label] = []

        self.stored_messages_for_local[label].append(messages)

    def get_messages_for_label(self, label):
        messages = []
        if label in self.stored_messages_for_local:
            messages = self.stored_messages_for_local[label].copy()
            self.stored_messages_for_local[label] = []
        return messages

    def any_messages_pending_for_label(self, label):
        if label not in self.stored_messages_for_local:
            return 0
        return len(self.stored_messages_for_local[label])

    @staticmethod
    def extract_english_japanese_text(frame_key, frame, ocr_config, label, ocr_key, regex=None):
#    def extract_english_japanese_text(self, frame):
        try:
            print('inside of extracto magneto')
            messages = PytesseractManager.process_image_for_text(frame, ocr_config)
        except Exception as e:
            print('Error occurred: %s ' % str(e))
            print(e)
            raise e
        return frame_key, messages, regex, label, ocr_key

    def construct_pokemon_info_page_1(self, future):
        frame_key, messages, regexes, label, ocr_key = future
        if frame_key not in self.processed_text:
            self.processed_text[frame_key] = {}

        if label not in self.processed_text[frame_key]:
            self.processed_text[frame_key][label] = []

        self.processed_text[frame_key][label].append(messages)

        self.messages_in_order.append((frame_key, label, messages))

        if regexes is None:
            self.add_messages_to_session(messages, label)
        else:
            for message in messages:
                for regex in regexes:
                    if re.search(regex, message):
                        if frame_key not in self.rejected_text:
                            self.rejected_text[frame_key] = {}

                        if label not in self.rejected_text[frame_key]:
                            self.rejected_text[frame_key][label] = []

                        self.rejected_text[frame_key][label].append(messages)
                        self.all_rejected_messages.append((frame_key, label, messages))
                        return
            self.add_messages_to_session(messages, label)
        self.all_accepted_messages.append((frame_key, label, messages))

        self.save_state(frame_key)


    @staticmethod
    def text_override_in_frame(img, rect_requirements, base_threshold):

        all_checks_pass = True
        failure_data = {}

        for requirement in rect_requirements:
            threshold = base_threshold
            if IGNORE_IF_TRIGGERS_KEY in requirement:
                threshold = requirement[THRESHOLD_KEY]

            crop_dimens = requirement[RECT_KEY]
            lower =(requirement["min_max_red"][0], requirement["min_max_green"][0], requirement["min_max_blue"][0]) # lower bound for each channel
            upper =(requirement["min_max_red"][1], requirement["min_max_green"][1], requirement["min_max_blue"][1]) # lower bound for each channel


            left = crop_dimens[0]
            top = crop_dimens[1]
            width = crop_dimens[2]
            height = crop_dimens[3]
            crop = img[top:height, left:width]

            all_checks_pass = all_checks_pass and (TextTracker.calculate_rect_ratio(crop, lower, upper) >= threshold)

            if not all_checks_pass:
                failure_data['crop_dimens'] = crop_dimens
                failure_data['requirement'] = requirement
                failure_data['percent'] = TextTracker.calculate_rect_ratio(crop, lower, upper)
                break

        return all_checks_pass, failure_data


    @staticmethod
    def calculate_rect_ratio(image, lower, upper):
        # make non grey pixels yellow
        img_adj = image.copy()

        # create the mask and use it to change the colors
        mask = cv2.inRange(img_adj, lower, upper)
        masked_color = [0,255,255]
        mask = 255 - mask
        img_adj[mask == 0] = masked_color
        return TextTracker.count_non_yellow_pixels(img_adj, masked_color)


    @staticmethod
    def count_non_yellow_pixels(image, mask):
        # grab the image dimensions
        h = image.shape[0]
        w = image.shape[1]
        target_pixels = 0


        # loop over the image, pixel by pixel
        for y in range(0, h):
            for x in range(0, w):
              pixel = image[y, x]
              if pixel[0] == mask[0] and pixel[1] == mask[1] and pixel[2] == mask[2]:
                target_pixels += 1
        return target_pixels / float(h*w)

    @staticmethod
    def calculate_similar_frame_ratio(imageData_1, imageData_2):

        valid_pixels_count = 0
        pixel_count = 0

        for i in range(0,imageData_1.shape[0]):
            for j in range(0,imageData_1.shape[1]):
                pixel_count += 1

                pixel_1 = [imageData_1.item(i, j, 0), imageData_1.item(i, j, 1), imageData_1.item(i, j, 2)]
                pixel_2 = [imageData_2.item(i, j, 0), imageData_2.item(i, j, 1), imageData_2.item(i, j, 2)]

                if pixel_1 == pixel_2:
                    valid_pixels_count += 1
        return valid_pixels_count / float(pixel_count)


    def save_message_img(self, message_frame, accepted):
        if not SAVE_MESSAGE_FRAMES:
            return

#        dir_to_use = '%s/%s' % (ACCEPTED_MESSAGE_DIR, self.state_manager.session_id)
        dir_to_use = ACCEPTED_MESSAGE_DIR
        if not accepted:
#            dir_to_use = '%s/%s' % (REJECTED_MESSAGE_DIR, self.state_manager.session_id)
            dir_to_use = REJECTED_MESSAGE_DIR

        if not os.path.exists(os.path.dirname(dir_to_use)):
            os.makedirs(os.path.dirname(dir_to_use))

        filename = '%s/%s.png' % (dir_to_use,  (str(uuid.uuid4())))
#        filename = '%s.png' % ((str(uuid.uuid4())))

        status = cv2.imwrite(filename, message_frame)


    @staticmethod
    def get_min_max_colors_for_regions_in_images(images, inspection_regions):

        result = []

        for crop_dimens in inspection_regions:
            left = crop_dimens[0]
            top = crop_dimens[1]
            width = crop_dimens[2]
            height = crop_dimens[3]

            black_min_red = 255
            black_max_red = 0
            black_min_green = 255
            black_max_green = 0
            black_min_blue = 255
            black_max_blue = 0

            for img in images:
                crop = img[top:height, left:width]

                # grab the image dimensions
                h = crop.shape[0]
                w = crop.shape[1]
                target_pixels = 0


                # loop over the image, pixel by pixel
                for y in range(0, h):
                    for x in range(0, w):

                        pixel = crop[y, x]
                        blue = pixel[0]
                        green = pixel[1]
                        red = pixel[2]

                        black_min_blue = int(min(black_min_blue, blue))
                        black_max_blue = int(max(black_max_blue, blue))
                        black_min_red = int(min(black_min_red, red))
                        black_max_red = int(max(black_max_red, red))
                        black_min_green = int(min(black_min_green, green))
                        black_max_green = int(max(black_max_green, green))

            result.append({
                "rect": crop_dimens,
                "min_max_red": [black_min_red, black_max_red],
                "min_max_green": [black_min_green, black_max_green],
                "min_max_blue": [black_min_blue, black_max_blue],
            })

        return result


    def compare_all_configs_againsts_other_images(self, config_mapper, restricted_labels = None):
        labels_to_track = self.config[LABELS_TO_TRACK_KEY]

        result = {}

        for label_to_track in labels_to_track:
            if restricted_labels is not None and label_to_track not in restricted_labels:
                continue

            tracking_configuration = self.config['tracking_configurations'][label_to_track]

            if 'override_rects' not in tracking_configuration:
                continue
            print("inspecting", label_to_track)

            threshold = tracking_configuration['threshold']

            other_images = set()

            for mapper_key in config_mapper["images"]:
                if mapper_key == label_to_track or config_mapper["images"][mapper_key] is None or len(config_mapper["images"][mapper_key]) == 0:
                    continue

                # removes lulua_item and lulua_atk combos
                for image_name in config_mapper["images"][mapper_key]:
                    if image_name not in config_mapper["images"][label_to_track]:
                        other_images.add(image_name)

            overriden_images = TextTracker.check_images_against_config_override(tracking_configuration, threshold, other_images)

            if len(overriden_images) > 0:
                result[label_to_track] = overriden_images

        return result

    # returns images that pass the override
    @staticmethod
    def check_images_against_config_override(self, tracking_configuration, threshold, image_names):
        overriden_images = []
        for image_name in image_names:
            image = cv2.imread(image_name)

            all_check_pass = True
            for rect_config in tracking_configuration['override_rects']:
                rect_image = TextTracker.get_subframe_from_frame(rect_config['rect'], image)
                # method expects list, since doing 1 at a time, force as list
                check_pass, failure_data = TextTracker.text_override_in_frame(image, [rect_config], threshold)
                all_check_pass = all_check_pass and check_pass
                if not all_check_pass:
                    break

            if all_check_pass:
                overriden_images.append(image_name)

        return sorted(overriden_images)


import json
import requests
import cv2
import numpy as np
import os


TESSERACT_BASE_URL = os.environ.get('TESSERACT_BASE_URL', 'http://localhost:5333/api/')
#TESSERACT_BASE_URL = os.environ.get('TESSERACT_BASE_URL', 'http://192.168.0.141:5333/api/')

class PytesseractManager():

    @staticmethod
    def fire_request( url, image, params):
        baseurl = TESSERACT_BASE_URL
        is_base64=False

        url = baseurl + url
        print('tesseract url:', url)

        if not is_base64:
#            headers = {'content-type': 'multipart/form-data'}

            _, image = cv2.imencode('.jpg', image)
            files = {
                'image': ('image', image, 'image/jpg'),
            }

            response = requests.post(url, data=params, files=files)

            return json.loads(response.text)['result']

        else:
            headers = {'content-type': 'application/json'}
            params['image'] =  image,

            response = requests.post(url, data=params, headers=headers)
            print('fire response', response)
            return json.loads(response.text)['result']

    @staticmethod
    def process_image_for_text( image, ocr_config=None):
        url = 'process_image_for_text'
        url = 'parse_rect_with_pytesseract_config'
        params = {}
        params['ocr_config'] = json.dumps(ocr_config)

        return PytesseractManager.fire_request(url, image, params)



import json
import enum

class METRIC_TYPES(enum.Enum):
    INCREMENT = 1
    MAX = 2
    APPEND = 3
    ADD_IF_DOESNT_EXIST = 4
    SET = 5
    SET_IF_NOT_NEGATIVE = 6
    ATLEAST_ONCE = 7
    MIN = 8

class MetricsTracker():
    def __init__(self, default_metrics):
        self.tracking_metrics = default_metrics
        self.metrics_this_turn = {}

        try:
            json.dumps(default_metrics)
        except Exception as e:
            print('default metrics is not serializable. if using a set, use list and use METRIC_TYPES.ADD_IF_DOESNT_EXIST to add items')
            raise e


    def bulk_update_metrics(self, metric_items):
        for prefix, metrics_key, value, metrics_type in metric_items:
            self.update_metrics(prefix, metrics_key, value, metrics_type)

    def update_metrics(self, prefix, metrics_key, value, metrics_type):
        full_key = metrics_key
        if metrics_key not in self.tracking_metrics.keys():
            full_key = '%s_%s' % (prefix, metrics_key)

#        print(full_key)
        if metrics_type == METRIC_TYPES.ATLEAST_ONCE:
            self.tracking_metrics[full_key] =  self.tracking_metrics[full_key] or value
        elif metrics_type == METRIC_TYPES.SET:
            self.tracking_metrics[full_key] = value
        elif metrics_type == METRIC_TYPES.MAX:
            self.tracking_metrics[full_key] = max(value, self.tracking_metrics[full_key])
        elif metrics_type == METRIC_TYPES.MIN:
            self.tracking_metrics[full_key] = min(value, self.tracking_metrics[full_key])
        elif metrics_type == METRIC_TYPES.APPEND:
            self.tracking_metrics[full_key].append(value)
        elif metrics_type == METRIC_TYPES.ADD_IF_DOESNT_EXIST:
            if value not in self.tracking_metrics[full_key]:
                self.tracking_metrics[full_key].append(value)
        elif metrics_type == METRIC_TYPES.INCREMENT:
            self.tracking_metrics[full_key] += 1
        elif metrics_type == METRIC_TYPES.SET_IF_NOT_NEGATIVE:
            if self.tracking_metrics[full_key] < 0:
                self.tracking_metrics[full_key] = value
        else:
            print('bad metrics_type', metrics_type)

        if full_key not in self.metrics_this_turn:
            self.metrics_this_turn[full_key] = []

        self.metrics_this_turn[full_key].append(self.tracking_metrics[full_key])


    def clear_metrics_this_turn(self):
        self.metrics_this_turn = {}



import cv2

stats_json = {
    'frame_count': 1,
    'assisted_count': 12,
    'assisted_filenames': [],
    'seen_labels': [],
    'assisted_labels': [],
    'filenames': [],
}

config_json = {
    'default_use_centroid': False,
    'centroid_labels': [],
    'rect_labels': [],
    'rect_percentage_threshold': 0.8,
    'centroid_distance_treshold': 50,
}

# raw image
IMAGES_DIR = './images'
#images with boxes
LABELED_IMAGES_DIR = './labeled_images'
# yolo txt files
YOLO_LABEL_DIR = './yolo'
#where metrics is
METRICS_DIR = './metrics'

import csv

import numpy as np
import json

import os
if not os.path.exists(IMAGES_DIR):
    os.makedirs(IMAGES_DIR)

if not os.path.exists(LABELED_IMAGES_DIR):
    os.makedirs(LABELED_IMAGES_DIR)

if not os.path.exists(YOLO_LABEL_DIR):
    os.makedirs(YOLO_LABEL_DIR)

if not os.path.exists(METRICS_DIR):
    os.makedirs(METRICS_DIR)

class AutoLabelTracker():
    def __init__(self, config, all_label_names, session_id):
        self.session_id = session_id
        self.config = config
        self.all_label_names  = all_label_names
        self.colors  = self.get_colors(all_label_names)

        self.first_of_kinds = []
        self.second_of_kinds = []

        self.frame_count = 0
        self.json_generator = []
        self.video_assets = {}
        self.image_asset_data = []
        self.last_label_keys = None

        self.frames_and_label_and_boxes = {}

        # For the sake of being about to update labels and boxes at a later time
        # and not having to loop through every source to fix reference
        self.labels_and_boxes_key_mapping = {}
        self.labels_to_ignore_during_comparison = []

        self.images_session_dir = '%s/%s' % (IMAGES_DIR, self.session_id)
        self.labeled_images_session_dir = '%s/%s' % (LABELED_IMAGES_DIR, self.session_id)
        self.yolo_label_session_dir = '%s/%s' % (YOLO_LABEL_DIR, self.session_id)
        self.metrics_session_dir = '%s/%s' % (METRICS_DIR, self.session_id)

        if not os.path.exists(self.images_session_dir):
            os.makedirs(self.images_session_dir)

        if not os.path.exists(self.labeled_images_session_dir):
            os.makedirs(self.labeled_images_session_dir)

        if not os.path.exists(self.yolo_label_session_dir):
            os.makedirs(self.yolo_label_session_dir)

        if not os.path.exists(self.metrics_session_dir):
            os.makedirs(self.metrics_session_dir)

    def is_same(self, label_and_boxes, other_label_and_boxes):
        rect_overlap_percent = 0.7
        label_keys = label_and_boxes.keys()
        other_label_keys = other_label_and_boxes.keys()

        # remove item from keys if in labels  blah blah b
        for ignored_label_key in self.labels_to_ignore_during_comparison:
            if ignored_label_key in label_keys:
                label_keys.remove(ignored_label_key)

            if ignored_label_key in other_label_keys:
                other_label_keys.remove(ignored_label_key)

        if len(set(label_keys)) != len(set(other_label_keys)):
            return False

        for label_key in label_keys:
            if label_key in self.labels_to_ignore_during_comparison:
                continue

            if label_key not in other_label_and_boxes:
                return False

            matches = []

            rects = label_and_boxes[label_key]
            other_rects = other_label_and_boxes[label_key]

            if len(rects) != len(other_rects):
                return False

            overlap_threshold = rect_overlap_percent

            for rect in rects:
                match_found = False
                for other_rect in other_rects:
                    if calculate_overlap(rect, other_rect) > overlap_threshold:
                        match_found = True
                        break

                if match_found == False:
                    return False

        return True


    def get_colors(self, LABELS):
        # initialize a list of colors to represent each possible class label
        np.random.seed(42)
        COLORS = np.random.randint(0, 255, size=(len(LABELS), 3),dtype="uint8")
        return COLORS

    def store_frame(self, original_filename, frame, label_and_boxes, assisted_labels, timestamp=None, video_filename=None):

        is_first_of_kind = False
        is_second_of_kind = False
        inside_of_first_of_a_kind = False
        for other_label_and_boxes in self.first_of_kinds:
            inside_of_first_of_a_kind = inside_of_first_of_a_kind or self.is_same(label_and_boxes, other_label_and_boxes)
            if inside_of_first_of_a_kind:
                break

        inside_of_second_of_a_kind = False
        for other_label_and_boxes in self.second_of_kinds:
            inside_of_second_of_a_kind = inside_of_second_of_a_kind or self.is_same(label_and_boxes, other_label_and_boxes)
            if inside_of_second_of_a_kind:
                break

        if not inside_of_first_of_a_kind and not inside_of_second_of_a_kind:
            self.first_of_kinds.append(label_and_boxes)
            is_first_of_kind = True

        elif not inside_of_second_of_a_kind:
            self.second_of_kinds.append(label_and_boxes)
            is_second_of_kind = True

        labels_and_boxes = []
        drawn_image = frame.copy()

        yolo_formats = []

        num_of_rects = 0
        # ensure at least one detection exists
        for label_name in label_and_boxes:
            rects = label_and_boxes[label_name]
            # loop over the indexes we are keeping
            for rect in rects:
                num_of_rects += 1
                # extract the bounding box coordinates
                (x1, y1) = (rect[0], rect[1])
                (x2, y2) = (rect[2], rect[3])

                color_index = self.all_label_names.index(label_name)

                # draw a bounding box rectangle and label on the image
                color = [int(c) for c in self.colors[color_index]]
                cv2.rectangle(drawn_image, (x1, y1), (x2, y2), color, 2)
                text = "{}: {:.4f}".format(label_name, 0)
                cv2.putText(drawn_image, text, (x1, y1 - 5), cv2.FONT_HERSHEY_SIMPLEX,0.5, color, 2)


            centroid, width, height = get_center_point_width_height(rect)
            xCenter, yCenter = centroid

            output = '%s %.6f %.6f %.6f %.6f\n' % (label_name, xCenter, yCenter, width, height)

            print(output)
            yolo_formats.append(output)

        filename = original_filename
        if len(assisted_labels) > 0:
            filename = 'assisted_%s' % original_filename

        dir_to_use = self.labeled_images_session_dir
        if not os.path.exists(os.path.dirname(dir_to_use)):
            os.makedirs(os.path.dirname(dir_to_use))

        write_filename = '%s/%s' % (dir_to_use, filename)

        status = cv2.imwrite(write_filename, drawn_image)

        dir_to_use = self.images_session_dir
        if not os.path.exists(os.path.dirname(dir_to_use)):
            os.makedirs(os.path.dirname(dir_to_use))
        write_filename = '%s/%s' % (dir_to_use, filename)

        status = cv2.imwrite(write_filename, frame)

        dir_to_use = self.yolo_label_session_dir
        filename_noext = filename.split('.')[0]
        textfile_out = '%s/%s.txt' % (dir_to_use, filename_noext)

        f = open(textfile_out, "w")
        f.writelines(yolo_formats)

        label_keys = label_and_boxes.keys()

        # check last frame for dramatic shift detection.
        dramatic_shift = False
        if self.last_label_keys is not None:
            if len(set(label_keys)) != len(set(self.last_label_keys)):
                dramatic_shift = True
        self.last_label_keys = label_keys


        csv_data = {
            'filename': original_filename,
            'assisted': len(assisted_labels) > 0,
            'assisted_labels': assisted_labels,
            'seen_labels': list(set(label_keys)),
            'num_of_rects': num_of_rects,
            'first_of_a_kind': is_first_of_kind,
            'second_of_a_kind': is_second_of_kind,
            "dramatic_shift": dramatic_shift,
        }
        if timestamp is not None:
            csv_data['timestamp'] = timestamp

        self.json_generator.append(csv_data)

        frame_str = str(original_filename)

        is_assisted = len(assisted_labels) > 0
        if timestamp is not None:
            if filename not in self.video_assets:
                self.video_assets[filename] = []
            self.video_assets[filename].append(
                    {
                        "filename": "",
                        "labels_and_boxes_key": frame_str,
                        "required_testing": is_second_of_kind,
                        "required_training": is_first_of_kind,
                        "is_assisted": is_assisted,
                        "dramatic_shift": dramatic_shift,
                        "timestamp": timestamp,
                    })
        else:
            self.image_asset_data.append({
                "filename": original_filename,
                "labels_and_boxes_key": frame_str,
                "required_testing": is_second_of_kind,
                "required_training": is_first_of_kind,
                "is_assisted": is_assisted,
                "dramatic_shift": dramatic_shift,
            })

        height, width, _ = frame.shape
        self.frames_and_label_and_boxes[frame_str] = {}
        self.frames_and_label_and_boxes[frame_str]['assisted'] = is_assisted
        self.frames_and_label_and_boxes[frame_str]['labels'] = label_and_boxes
        self.frames_and_label_and_boxes[frame_str]['artifacts_filename'] = filename
        self.frames_and_label_and_boxes[frame_str]['manual'] = False
        self.frames_and_label_and_boxes[frame_str]['width'] = width
        self.frames_and_label_and_boxes[frame_str]['height'] = height
        self.frames_and_label_and_boxes[frame_str]['required_testing'] = is_second_of_kind
        self.frames_and_label_and_boxes[frame_str]['required_training'] = is_first_of_kind

        self.labels_and_boxes_key_mapping[frame_str] = label_and_boxes

    def wrapup(self):
        print('writing metrics')
        self.metrics_session_dir = '%s/%s' % (METRICS_DIR, self.session_id)

        with open('%s/%s' % (self.metrics_session_dir, 'metrics.json'), "w") as outfile:
            json.dump(self.json_generator, outfile, indent=4)

        vott_csv_filename = '%s_vott_export.csv' % (self.session_id)
        vott_csv_filename = 'vott_export.csv'

        with open('%s/%s' % (self.metrics_session_dir, vott_csv_filename), "w") as csvfile:
            # creating a csv writer object
            csvwriter = csv.writer(csvfile)

            # writing the fields
            csvwriter.writerow(["image","xmin","ymin","xmax","ymax","label", "width", "height"])

            for filename_key in self.frames_and_label_and_boxes:
                output_data = self.frames_and_label_and_boxes[filename_key]

                for output_label in output_data['labels']:
                    for output_rect in output_data['labels'][output_label]:
                        imagename = output_data['artifacts_filename']
                        xmin = output_rect[0]
                        ymin = output_rect[1]
                        xmax = output_rect[2]
                        ymax = output_rect[3]
                        label = output_label
                        width = output_data['width']
                        height = output_data['height']

                        csvwriter.writerow([imagename, xmin, ymin, xmax, ymax, label, width, height])

        required_testing_filename = '-testing-export.txt'
        required_testing_filename = '%s_testing-export.txt' % (self.session_id)
        with open('%s/%s' % (self.metrics_session_dir, required_testing_filename), "w") as csvfile:
            # creating a csv writer object
            csvwriter = csv.writer(csvfile)

            # writing the fields
            csvwriter.writerow(["name"])

            for filename_key in self.frames_and_label_and_boxes:
                output_data = self.frames_and_label_and_boxes[filename_key]

                if output_data['required_testing']:
                    csvwriter.writerow([output_data['artifacts_filename']])


        required_training_filename = '-training-export.txt'
        required_training_filename = '%s_training-export.txt' % (self.session_id)
        with open('%s/%s' % (self.metrics_session_dir, required_training_filename), "w") as csvfile:
            # creating a csv writer object
            csvwriter = csv.writer(csvfile)

            # writing the fields
            csvwriter.writerow(["name"])

            for filename_key in self.frames_and_label_and_boxes:
                output_data = self.frames_and_label_and_boxes[filename_key]

                if output_data['required_training']:
                    csvwriter.writerow([output_data['artifacts_filename']])




    # ignore processed/rejected text should always be saved elsewhere
    def update_state(self, frame_key, label_and_boxes):
        self.labels_and_boxes_key_mapping[frame_key] = label_and_boxes
        self.frames_and_label_and_boxes[frame_key]['manual'] = True
        self.frames_and_label_and_boxes[frame_key]['labels'] = label_and_boxes


def get_center_point_width_height(rect):
	width = rect[2] - rect[0]
	height = rect[3] - rect[1]
	startX, startY, endX, endY = rect
	cX = int((startX + endX) / 2.0)
	cY = int((startY + endY) / 2.0)

	return [cX, cY], width, height

def distance_between_points(p1, p2):
	return math.sqrt( ((p1[0]-p2[0])**2)+((p1[1]-p2[1])**2) )

def calculate_overlap(rect_1, rect_2):
    XA2 = rect_1[2]
    XB2 = rect_2[2]
    XA1 = rect_1[0]
    XB1 = rect_2[0]

    YA2 = rect_1[3]
    YB2 = rect_2[3]
    YA1 = rect_1[1]
    YB1 = rect_2[1]
    # overlap between A and B
    SA = (rect_1[2] - rect_1[0]) * (rect_1[3] - rect_1[1])
    SB = (rect_2[2] - rect_2[0]) * (rect_2[3] - rect_2[1])

    SI = max(0, 1+ min(XA2, XB2) - max(XA1, XB1)) * max(0, 1 + min(YA2, YB2) - max(YA1, YB1))
    SU = SA + SB - SI
    if float(SA) == 0.0:
        return 0
    return SI/float(SA)
